// Author K.Karthik
// Version 1.0
// Category Freeware (Non-Visual delphi4 Unit)
// Notes:
// I wrote this unit, because i couldn't find any such unit in
// Delphi Components Page
// Shows Imported functions,Exported Functions and Sections names
// in a PE
// Portable Executable(PE) Format Dump
// Supports Direct Disk Read(Faster, if file is in CD)
// and also Memory Read(faster, if file is in HDD)
// If you are using this unit or updating it Send a mail
// to karthik_prg@yahoo.com
// Steps:
// 1. Always Call PEHDR.Create first.
// 2. Set PEHDR.LoadinMemory to True if the File is to mapped in Memory
//    else set it false
// 3. Call PEHDR.Load(filename) and test if it is successful
// 4. Call PEHDR.resolve and test if it is successful
// 5. if no errors, uses imported, Exported and SectionNames Tstrings
// 6. Call PEHDR.Clear
// 7. Call PEHDR.Free

unit pe;
interface

uses windows,classes,sysutils;

const IMAGE_ORDINAL_FLAG32:DWORD=$80000000;

type IMAGE_DOS_HEADER=record
  e_magic:WORD;
  e_cblp:WORD;
  e_cp:WORD;
  e_cric:WORD;
  e_cparhdr:WORD;
  e_minalloc:WORD;
  e_maxalloc:WORD;
  e_ss:WORD;
  e_sp:WORD;
  e_csum:WORD;
  e_ip:WORD;
  e_cs:WORD;
  e_lfarlc:WORD;
  e_ovno:WORD;
  e_res:array[0..3] of WORD;
  e_oemid:WORD;
  e_oeminfo:WORD;
  e_res2:array[0..9] of WORD;
  e_lfanew:DWORD;
end;

type Dosret=record
 NThdr:DWORD;
 valid:Boolean;
end;

type PIMAGE_NT_HEADERS=^IMAGE_NT_HEADERS;
type PIMAGE_SECTION_HEADER=^IMAGE_SECTION_HEADER;

type winret=record
 NTHdr:PIMAGE_NT_HEADERS;
 valid:Boolean;
end;

type IMAGE_IMPORT_DESCRIPTOR=record
 OriginalFirstThunk:DWORD;
 TimeDateStamp:DWORD;
 ForwarderChain:DWORD;
 Name:DWord;
 FirstThunk:DWORD;
end;

type IMAGE_IMPORT_BY_NAME=record
 Hint:WORD;
 Name:Byte;
end;

type IMAGE_EXPORT_DIRECTORY =record
    Characteristics:DWORD;
    TimeDateStamp:DWORD;
    MajorVersion:WORD;
    MinorVersion:WORD;
    Name:DWORD;
    Base:DWORD;
    NumberOfFunctions:DWORD;
    NumberOfNames:DWORD;
    AddressOfFunctions:DWORD;
    AddressOfNames:DWORD;
    AddressOfNameOrdinals:DWORD;
end;

type PEHDR=class(TComponent)
private
    IFname:string;
    NTStream:Tstream;
    function verifyDosHeader(F:Tstream):Dosret;
    function verifyPEheader(F:Tstream;offset:DWORD):winret;
    function GetLinearAddress(Sh:PIMAGE_SECTION_HEADER;count:DWORD;rva:DWORD):DWORD;
    procedure LoadImports(f:Tstream;offset:DWORD;PSection:PIMAGE_SECTION_HEADER;SectionC:DWORD;return:Tstrings);
    procedure LoadExports(f:Tstream;offset:DWORD;PSection:PIMAGE_SECTION_HEADER;SectionC:DWORD;return:Tstrings);
    Procedure LoadSectionNames(f:Tstream;p:PIMAGE_SECTION_HEADER;count:DWORD;names:Tstrings);
    function LoadSectionData(f:Tstream;offset,count:DWORD):Pointer;

public
    imported:Tstrings;
    Exported:Tstrings;
    SectionNames:Tstrings;
    Filename:string;
    LoadinMemory:Boolean;
    function Load(Fname:String):Boolean;
    function Resolve:Boolean;
    Procedure Clear;

end;

implementation

function PEHDR.Load(Fname:String):Boolean;
var
 Flag:Boolean;
 hFile,Fmap:THandle;
 Ptr:Pointer;
 tmp:TfileStream;
Begin
 Filename:=Fname;
 Flag:=True;
 try
    if(LoadinMemory) then
    Begin
      tmp:=TFilestream.Create(Fname,fmOpenRead or fmShareCompat or fmShareDenyNone);
      NTStream:=TMemoryStream.Create;
      TMemoryStream(NTStream).LoadFromStream(tmp);
      tmp.Free;
    end
    else
      NTstream:=TFilestream.Create(Fname,fmOpenRead or fmShareCompat or fmShareDenyNone);
 except on Exception do Begin Flag:=False; end;
 end;
 if(not Flag) then begin result:=False; exit; end;
 imported:=Tstringlist.Create;
 Exported:=Tstringlist.Create;
 SectionNames:=Tstringlist.Create;
 result:=True;
end;

function PEHDR.Resolve:Boolean;
var
 Dret:Dosret;
 NTret:winret;
 Sret:Pointer;
Begin
 Dret:=verifyDosHeader(NTStream);
 if(not Dret.valid) then Begin Result:=False; exit; end;
 NTret:=verifyPEHeader(NTStream,Dret.NTHdr);
 if(not NTret.valid) then Begin Result:=False; exit; end;
 Sret:=LoadSectionData(NTStream,Dret.NTHdr,NTret.NTHdr^.FileHeader.NumberofSections);
 LoadSectionNames(NTStream,Sret,NTret.NTHdr^.FileHeader.NumberofSections,SectionNames);
 LoadImports(NTStream,DRet.NTHdr,Sret,NTret.NTHdr^.FileHeader.NumberofSections,imported);
 LoadExports(NTStream,DRet.NTHdr,Sret,NTret.NTHdr^.FileHeader.NumberofSections,Exported);
 if(Sret<>Nil) then
    Freemem(Sret);
end;

Procedure PEHDR.Clear;
Begin
 imported.Clear;
 Exported.Clear;
 SectionNames.Clear;
 imported.Free;
 Exported.Free;
 SectionNames.Free;
 NTStream.Free;
end;

function PEHDR.verifyDosHeader(F:Tstream):Dosret;
var
 DHdr:IMAGE_DOS_HEADER;
 count:integer;
 ret:Dosret;
begin
 F.Seek(0,soFromBeginning);
 count:=F.Read(DHdr,sizeof(DHdr));
 if((count<sizeof(Dhdr)) or (Dhdr.e_magic<>$5a4d)) then
  ret.valid:=False
 else
 Begin
  ret.NTHdr:=DHdr.e_lfanew;
  ret.valid:=True;
 end;
 result:=ret;
end;

function PEHDR.verifyPEheader(F:Tstream;offset:DWORD):winret;
var
  ret:winret;
  count:integer;
begin
 new(ret.NTHdr);
 F.Seek(offset,soFromBeginning);
 count:=F.Read(ret.NTHdr^,sizeof(IMAGE_NT_HEADERS));
 if((count<>sizeof(ret.NTHdr^)) or (ret.NTHdr^.Signature<>$4550)) then
  ret.valid:=False
 else
  ret.valid:=True;
 Result:=Ret;
end;


function PEHDR.GetLinearAddress(Sh:PIMAGE_SECTION_HEADER;count:DWORD;rva:DWORD):DWORD;
var
 i,t:integer;
 st:string;
Begin
 t:=0;
 Result:=0;
 while( t < count) do
 Begin
   if((rva >= Sh^.VirtualAddress) and
     (rva<(Sh^.VirtualAddress+Sh^.SizeofRawData))) then
    Begin
      Result:=rva+Sh^.PointertoRawData-Sh^.VirtualAddress;
      break;
    end;
   inc(t);
  Sh:=Pointer(DWORD(Sh)+sizeof(IMAGE_SECTION_HEADER));
  end;
end;

procedure PEHDR.LoadExports(f:Tstream;offset:DWORD;PSection:PIMAGE_SECTION_HEADER;SectionC:DWORD;return:Tstrings);
var
 NT:IMAGE_NT_HEADERS;
 Exp:IMAGE_EXPORT_DIRECTORY;
 off,AddrN,i,tmp,LAddrN:DWORD;
 wd:WORD;
 ch:Char;
 S:String;
 Names:^DWORD;
begin
 f.Seek(offset,soFromBeginning);
 f.read(NT,sizeof(NT));
 if(NT.OptionalHeader.DataDirectory[0].virtualAddress=0) then
 Begin
   Exported.Clear;
   exit;
 end;
 off:=GetLinearAddress(PSection,SectionC,NT.OptionalHeader.DataDirectory[0].virtualAddress);
 f.seek(off,soFromBeginning);
 f.read(Exp,sizeof(Exp));
 AddrN:=GetLinearAddress(PSection,SectionC,Exp.AddressofNames);
 Names:=AllocMem(4*Exp.NumberofNames);
 f.seek(AddrN,soFromBeginning);
 f.Read(Names^,4*Exp.NumberofNames);
 for i:=1 to Exp.NumberofNames do
 Begin
    LAddrN:=GetLinearAddress(PSection,SectionC,Names^);
    f.Seek(LAddrN,soFromBeginning);
    S:='';
    Ch:='a';
    while(ord(ch)<>0) do
    Begin
      f.Read(Ch,sizeof(Char));
      S:=S+Ch;
    end;
    Exported.Add(S);
    inc(Names);
 end;
end;

procedure PEHDR.LoadImports(f:Tstream;offset:DWORD;PSection:PIMAGE_SECTION_HEADER;SectionC:DWORD;return:Tstrings);
var
 NT:IMAGE_NT_HEADERS;
 import:IMAGE_IMPORT_DESCRIPTOR;
 off,Thunk:DWORD;
 imp:IMAGE_IMPORT_BY_NAME;
 S:String;
 Ch:Char;
 imgoffset,addr,pos1:DWORD;
 ord1:Dword;
 hint:word;
Begin
 f.Seek(offset,soFromBeginning);
 f.read(NT,sizeof(NT));
 off:=GetLinearAddress(PSection,SectionC,NT.OptionalHeader.DataDirectory[1].virtualAddress);
 f.seek(off,soFromBeginning);
 f.read(import,sizeof(import));
 while (not((import.OriginalFirstThunk=0) and (import.FirstThunk=0)
            and (import.TimeDateStamp=0) and (import.Name=0)
            and (import.ForwarderChain=0))) do
 Begin
 f.Seek(GetLinearAddress(PSection,SectionC,import.Name),soFromBeginning);

 S:='';
 Ch:='a';
 while(ord(Ch)<>0) do
 Begin
  f.Read(Ch,sizeof(Char));
  S:=S+Ch;
 end;
 return.add(S);
 return.add('');
 Thunk:=import.OriginalFirstThunk;
 if (Thunk=0) then
   Thunk:=import.FirstThunk;
 f.Seek(GetLinearAddress(PSection,SectionC,Thunk),soFromBeginning);
 f.read(addr,sizeof(DWORD));
 Pos1:=f.Position;
 while(addr<>0) do
 Begin
    if((addr or IMAGE_ORDINAL_FLAG32)=addr) then
    Begin
        return.add(inttostr(lo(addr))+'    (ordinal)');
    end else
    Begin
        f.Seek(getLinearAddress(PSection,SectionC,addr),soFromBeginning);
        f.read(hint,sizeof(word));
        Ch:='a';
        S:='';
        while(ord(Ch)<>0) do
        Begin
            f.Read(Ch,sizeof(Char));
            S:=S+Ch;
        end;
        return.add(inttostr(hint)+'    '+S);
    end;
    f.Seek(Pos1,SoFromBeginning);
    f.read(addr,sizeof(DWORD));
    Pos1:=f.Position;
 end;
 off:=off+sizeof(IMAGE_IMPORT_DESCRIPTOR);
 f.seek(off,soFromBeginning);
 f.read(import,sizeof(import));
 return.add('');
 return.add('');
 end;
end;

Procedure PEHDR.LoadSectionNames(f:Tstream;p:PIMAGE_SECTION_HEADER;count:DWORD;names:Tstrings);
var
 s:string;
 i,k:integer;
Begin
 k:=0;
 while(k<count) do
 Begin
   s:='';
   for i:=0 to 7 do
    s:=s+char(p^.Name[i]);
   names.add(s);
   p:=Pointer(DWORD(p)+sizeof(IMAGE_SECTION_HEADER));
   inc(k);
 end;
end;

function PEHDR.LoadSectionData(f:Tstream;offset,count:DWORD):Pointer;
var
 ret:Pointer;
 Fret:Pointer;
 i:integer;
Begin
 f.Seek(offset+sizeof(IMAGE_NT_HEADERS),soFromBeginning);
 ret:=Allocmem(sizeof(IMAGE_SECTION_HEADER)*count);
 if(ret<>Nil) then
 f.Read(ret^,sizeof(IMAGE_SECTION_HEADER)*count);
 result:=ret;
end;

end.


